package dk.sdu.imada.jkkn04.Protein.data;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;

/**
 * @author Kristian Krmmer Nielsen <jkkn@jkkn.dk>
 */
public class Protein {

	private static final Pattern IDENTIFIER_PATTERN = Pattern
			.compile("^([A-Z0-9]{4})[:_]?([A-Z0-9]?).*");
	private final String description;
	private final String identifier;
	private final String rawSequence;
	private final String sequence;
	private String cleanSequence;
	private final SequenceType sequenceType;
	private final TranslationType translationType;

	/**
	 * Represents type of translation
	 */
	public enum TranslationType {
		/**
		 * Raw sequence, as it was read
		 */
		RAW,
		/**
		 * Normal translation ("b." mapped to "EC")
		 */
		NORMAL,
		/**
		 * STRIDE/DSSP ("GHIEBTS.-" mapped to "HHHEECCCC")
		 */
		TRANS_A,
		/**
		 * STRIDE/DSSP ("HEGITSB.-" mapped to "HECCCCCCC")
		 */
		TRANS_B
	}

	/**
	 * Creates new protein sequence
	 * 
	 * @param description Description for protein sequence
	 * @param rawSequence Raw sequence string
	 * @param predictionType Type of prediction/observation
	 * @param translationType Translation to use
	 */
	public Protein(final String description, final String rawSequence,
			final SequenceType predictionType, final TranslationType translationType) {
		super();
		this.description = description.trim();
		identifier = cleanIdentifer();

		this.rawSequence = rawSequence;
		sequenceType = predictionType;
		this.translationType = translationType;

		sequence = translatedProtein();
		cleanSequence = null;
	}

	private String cleanIdentifer() {
		/* Clean identifier similar to how Fiona does it */
		final Matcher XXXX_X = IDENTIFIER_PATTERN.matcher(description.toUpperCase());
		String clean = "";

		if (XXXX_X.matches()) {
			clean = XXXX_X.group(1) + "_" + XXXX_X.group(2);
			if (XXXX_X.group(2).length() == 0) {
				clean += "_"; // add anonymous chain identifier
			}
		} else {
			clean = "BAD_NAME_FORMAT: " + description;
			// throw something
		}

		return clean;
	}

	/**
	 * Returns description of protein sequence
	 * 
	 * @return description
	 */
	public String getDescription() {
		return description;
	}

	/**
	 * Returns identifier for protein sequence
	 * 
	 * @return identifier
	 */
	public String getIdentifier() {
		return identifier;
	}

	/**
	 * Returns translated sequence
	 * 
	 * @return translated sequence
	 */
	public String getSequence() {
		return sequence;
	}

	private String translatedProtein() {
		String sequence = rawSequence;
		final String upperCase = rawSequence.toUpperCase();

		switch (translationType) {
		case NORMAL:
			sequence = StringUtils.replaceChars(rawSequence, "b.", "EC");
			break;
		case TRANS_A:
			sequence = StringUtils.replaceChars(upperCase, "GHIEBTS.-", "HHHEECCCC");
			break;
		case TRANS_B:
			sequence = StringUtils.replaceChars(upperCase, "HEGITSB.-", "HECCCCCCC");
			break;
		case RAW:
		default:
			break;
		}

		return sequence;
	}

	/**
	 * Creates a list of possible translation of the given protein
	 * 
	 * @param p Protein to translate
	 * @return list of translations
	 */
	public static List<Protein> getTranslatedProteins(final Protein p) {
		final List<Protein> list = new ArrayList<Protein>(2);
		if (p.sequenceType.hasExtendedChars()) {
			list.add(new Protein(p.description, p.rawSequence, p.sequenceType,
					TranslationType.TRANS_A));
			list.add(new Protein(p.description, p.rawSequence, p.sequenceType,
					TranslationType.TRANS_B));
		} else if (!p.sequenceType.isPrimary()) {
			list.add(new Protein(p.description, p.rawSequence, p.sequenceType,
					TranslationType.NORMAL));
		} else {
			list.add(p);
		}
		return list;
	}

	/**
	 * Returns a cleaned sequence (X in reference sequence is instead in current sequence)
	 * 
	 * @return cleaned sequence
	 */
	public String getCleanSequence() {
		if (cleanSequence == null)
			return sequence;
		return cleanSequence;
	}

	/**
	 * Returns type of current sequence
	 * 
	 * @return type of sequence
	 */
	public SequenceType getSequenceType() {
		return sequenceType;
	}

	/**
	 * Sets a reference sequence. This is used to filter out unknown. Unknowns (X) in
	 * reference sequence replaces anything in the same location in the cleaned version of
	 * the current sequence.
	 * 
	 * @param reference Sequence to use as reference
	 */
	public void setReferenceSequence(final Protein reference) {
		cleanProtein(reference);
	}

	private void cleanProtein(final Protein reference) {
		final String refSeq = reference.getSequence();

		// Find positions of unknowns.
		final List<Integer> positions = new ArrayList<Integer>();
		int unknownPos = 0;
		while ((unknownPos = refSeq.indexOf('X', unknownPos)) != -1) {
			positions.add(unknownPos++);
		}
		if (!sequenceType.isObservation()) {
			Collections.reverse(positions); // go backwards thru list
		}

		final StringBuilder cleanSeq = new StringBuilder(sequence);
		for (final Integer pos : positions) {
			if (!sequenceType.isObservation()) {
				/*
				 * we remove unknowns inserting X's at their position appropriately as
				 * they appear in the reference
				 */
				cleanSeq.setCharAt(pos, 'X');
			} else {
				/*
				 * we will insert X's appropriately as they appear in the reference
				 */
				cleanSeq.insert((int) pos, 'X');
			}
		}

		cleanSequence = cleanSeq.toString();
	}

	/**
	 * Return translated cleaned sequence without unknowns.
	 * 
	 * @return sequence where X's are removed.
	 */
	public String getSequenceWithoutUnknowns() {
		return getCleanSequence().replace("X", "");
	}

	/**
	 * Returns type of sequence
	 * 
	 * @return sequence type
	 */
	public TranslationType getTranslationType() {
		return translationType;
	}

}
