package dk.sdu.imada.jkkn04.Protein.tests;

import java.util.Map;
import java.util.TreeMap;

import dk.sdu.imada.jkkn04.Protein.data.Protein;

/**
 * This abstract class can be implemented to offer a new evaluation type for comparing a
 * predicted secondary protein sequence with a observed sequence.
 * 
 * @author Kristian Krmmer Nielsen <jkkn@jkkn.dk>
 */
public abstract class AbstractEvaluation implements Cloneable {

	/**
	 * Observed protein sequence to use for evaluation
	 */
	protected Protein observedProtein;
	/**
	 * Predicted protein sequence to use for evaluation
	 */
	protected Protein predictedProtein;

	/**
	 * Represents the different states in a protein sequence
	 */
	public enum State {
		/**
		 * Helix
		 */
		H,
		/**
		 * Strand
		 */
		E,
		/**
		 * Coil
		 */
		C;

		/**
		 * Total number of different states we handle.
		 */
		final static int COUNT = State.values().length;

		/**
		 * Character representing this state
		 * 
		 * @return char (H/E/C)
		 */
		public char getChar() {
			return name().charAt(0);
		}

		/**
		 * Number representing this state
		 * 
		 * @return number between -1 and 2 (-1 in case of X)
		 */
		public int getInt() {
			return mapCharToInt(getChar());
		}
	};

	/**
	 * Number of sequences considered in this evaluation
	 */
	protected int sequences = 1;

	/**
	 * Returns results for this evaluation.<br/>
	 * Should be overridden to include the actual results
	 * 
	 * @return Map of key/value of results
	 */
	public Map<String, Object> getResults() {
		final Map<String, Object> results = new TreeMap<String, Object>();
		results.put("sequences", sequences);
		return results;
	}

	/**
	 * Method for translating a character (H/E/C/X) into a number
	 * 
	 * @param c H/E/C/X state
	 * @return number between -1 and 2 (-1 in case of X)
	 */
	public static int mapCharToInt(final char c) {
		switch (c) {
		case 'H':
			return 0;
		case 'E':
			return 1;
		case 'C':
			return 2;
		case 'X':
			return -1;
		default:
			System.err.println("Bad character: " + c);
		}
		return -1;
	}

	/**
	 * Method for translating number into a character (H/E/C/X)
	 * 
	 * @param i Number between -1 and 2.
	 * @return H/E/C/X.
	 */
	public static char mapIntToChar(final int i) {
		switch (i) {
		case 0:
			return 'H';
		case 1:
			return 'E';
		case 2:
			return 'C';
		case -1:
			return 'X';
		default:
			System.err.println("Bad integer: " + i);
		}
		return '?';
	}

	/**
	 * Configures the evaluator with the observed protein sequence to use
	 * 
	 * @param observedProtein observed sequence
	 */
	public void setObservedProtein(final Protein observedProtein) {
		this.observedProtein = observedProtein;
	}

	/**
	 * Configures the evaluator with the prediction protein sequence to use
	 * 
	 * @param predictedProtein prediction sequence
	 */
	public void setPredictedProtein(final Protein predictedProtein) {
		this.predictedProtein = predictedProtein;
	}

	/**
	 * Main method, must be implemented to perform the evaluation
	 */
	abstract public void performEvaluation();

	/**
	 * Creates a clone of the result, useful when merging results
	 * 
	 * @see java.lang.Object#clone()
	 */
	@Override
	public AbstractEvaluation clone() {
		try {
			final AbstractEvaluation clone = (AbstractEvaluation) super.clone();
			clone.observedProtein = clone.predictedProtein = null;
			return clone;
		} catch (final CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * Sums results from this evaluation with results for another sequence
	 * 
	 * @param evaluation Another evaluation of same type
	 */
	public void sumWith(final AbstractEvaluation evaluation) {
		sequences += evaluation.sequences;
	}

}