package dk.sdu.imada.jkkn04.Protein.tests;

import java.text.CharacterIterator;
import java.text.StringCharacterIterator;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author Kristian Krmmer Nielsen <jkkn@jkkn.dk>
 */
public class StateMachine extends AbstractEvaluation {

	/**
	 * Set of different confusion parameter which we measure
	 */
	public enum ConfusionParameter {
		/**
		 * In case that a predictor have made a state longer than it should be
		 */
		EXPANSION,
		/**
		 * In case that a predictor have made a state shorter than it should be
		 */
		REDUCTION,
		/**
		 * In case a predictor kept state and left out a state change
		 */
		BRIDGE,
		/**
		 * In case a predictor changes state in the middle but returned to the correct
		 * state afterwards
		 */
		GAP,
		/**
		 * In case of any other mismatch
		 */
		OTHER;
	}

	/**
	 * This represents the confusion matrix for one confusion parameter.
	 */
	protected class ConfusionCount {
		/**
		 * Number of errors seen from one state to another.
		 */
		public int matrix[][] = new int[State.COUNT][State.COUNT];
	}

	private enum MachineState {
		CORRECT, MISMATCH, UNKNOWNS, END
	}

	/**
	 * Inner class used to represent location in a sequence (predicted or observed)
	 */
	protected class SequenceCounter {
		
		
		/**
		 * Create new Sequence counter for a given sequence string
		 * @param sequence sequence
		 */
		public SequenceCounter(String sequence) {
			this.iterator = new StringCharacterIterator(sequence);
		}

		/**
		 * Iterator used for iterating by each character of the sequence
		 */
		CharacterIterator iterator;

		/**
		 * Is set to true if a previous valid character was repeated in an invalid
		 * position
		 */
		boolean hasRepeated = false;

		/**
		 * Character which was repeated
		 */
		char repeated;

		/**
		 * Number of times character was repeated since last valid match
		 */
		int countRepeated = 0;

		/**
		 * Repeated character is linked all the may to last valid match
		 */
		boolean linksToCorrect = false;
	};

	private int sequenceLength;
	private Map<ConfusionParameter, ConfusionCount> confusionMatrix = null;

	/* used while processing */
	private SequenceCounter observ;
	private SequenceCounter predict;

	/* Mismatches registered on sequence */
	private List<Set<ConfusionParameter>> mismatchesOnSeq;

	/**
	 * Performs the calculation of the State-matrix
	 * 
	 * @see dk.sdu.imada.jkkn04.Protein.tests.AbstractEvaluation#performEvaluation()
	 */
	@Override
	public void performEvaluation() {
		final String seqObserved = observedProtein.getCleanSequence();
		final String seqPredict = predictedProtein.getCleanSequence();

		if (seqObserved.length() != seqPredict.length())
			// FIXME: throw something
			return;
		sequenceLength = seqObserved.length();

		/* Initialise matrix */
		confusionMatrix = new EnumMap<ConfusionParameter, ConfusionCount>(
				ConfusionParameter.class);
		for (final ConfusionParameter p : ConfusionParameter.values()) {
			confusionMatrix.put(p, new ConfusionCount());
		}

		observ  = new SequenceCounter(seqObserved);
		predict = new SequenceCounter(seqPredict);

		/* No mismatches found yet, set all sets to empty */
		mismatchesOnSeq = new ArrayList<Set<ConfusionParameter>>(sequenceLength);
		for (int i = 0; i < sequenceLength; i++) {
			mismatchesOnSeq.add(EnumSet.noneOf(ConfusionParameter.class));
		}

		int mismatches = 0;

		MachineState currentState;
		char chrObserved = observ.iterator.current();
		char chrPredict = predict.iterator.current();

		do {
			if (chrObserved == 'X' || chrPredict == 'X') {
				// skipping unknowns / processed as end of string
				currentState = MachineState.UNKNOWNS;
			} else if (chrObserved == chrPredict) {
				if (chrObserved == CharacterIterator.DONE
						|| chrPredict == CharacterIterator.DONE) {
					currentState = MachineState.END;
				} else {
					currentState = MachineState.CORRECT;
				}
			} else {
				currentState = MachineState.MISMATCH;
			}

			/*
			 * Count possible gaps / reductions seeing them from the observed sequence
			 */
			processSeenFromSequence(currentState, observ, ConfusionParameter.GAP,
					ConfusionParameter.REDUCTION);
			/*
			 * Count possible bridges / expansion seeing them from the predicted sequence
			 */
			processSeenFromSequence(currentState, predict, ConfusionParameter.BRIDGE,
					ConfusionParameter.EXPANSION);

			switch (currentState) {
			case END:
			case CORRECT:
			case UNKNOWNS:
				if (mismatches > 0) {
					registerOther(mismatches);
					mismatches = 0;
				}
				break;
			case MISMATCH:
				mismatches++;
				break;
			}

			if (currentState != MachineState.END) {
				chrObserved = observ.iterator.next();
				chrPredict = predict.iterator.next();
			}

		} while (currentState != MachineState.END);

	}

	private void processSeenFromSequence(final MachineState currentState,
			final SequenceCounter sequence, final ConfusionParameter bridge_or_gap,
			final ConfusionParameter expansion_or_reduction) {
		final char chr = sequence.iterator.current();
		switch (currentState) {
		case END:
		case CORRECT:
		case UNKNOWNS:
			if (sequence.hasRepeated && sequence.countRepeated > 0) {
				if (sequence.linksToCorrect && sequence.repeated == chr) {
					/*
					 * BRIDGE (linked all the way in prediction and still same state) or
					 * GAP (linked all the way in observation and still same state)
					 */
					registerMismatch(bridge_or_gap, sequence.countRepeated);
				} else if (sequence.linksToCorrect || sequence.repeated == chr) {
					/*
					 * EXPANSION (if repeated previous prediction or repeated current
					 * prediction) or
					 * REDUCTION (if repeated previous observation or repeated
					 * current observation)
					 */
					registerMismatch(expansion_or_reduction, sequence.countRepeated);
				}
				sequence.countRepeated = 0;
			}
			/*
			 * MATCH: store information about last matched character and mark as linked to
			 * correct char
			 */
			if (currentState == MachineState.CORRECT) {
				sequence.repeated = chr;
				sequence.linksToCorrect = true;
				sequence.hasRepeated = true;
			} else {
				/* else we are at end of string/segment */
				sequence.linksToCorrect = sequence.hasRepeated = false;
			}

			break;
		case MISMATCH:
			if (sequence.hasRepeated && sequence.repeated == chr) {
				sequence.countRepeated++;
			} else {
				if (sequence.linksToCorrect) {
					/* EXPANSION or REDUCTION (notice countRepeated may be zero) */
					registerMismatch(expansion_or_reduction, sequence.countRepeated);
				}
				sequence.linksToCorrect = false;
				sequence.countRepeated = 1;
			}
			sequence.repeated = chr;
			sequence.hasRepeated = true;
			break;
		}
	}

	private void registerOther(final int mismatches) {
		/*
		 * We look if we have mismatching entries which are not registered under any type
		 * if so, we will count the mismatches under OTHER
		 */

		/* rewind iterators */
		observ.iterator.setIndex(observ.iterator.getIndex() - mismatches);
		predict.iterator.setIndex(predict.iterator.getIndex() - mismatches);

		/* update other counts */
		final ConfusionCount m = confusionMatrix.get(ConfusionParameter.OTHER);

		/* count as "other" mismatch while no bridge/gap/reduction/expansion registered */
		for (int x = 0, index = observ.iterator.getIndex(); x < mismatches; x++, index++) {
			final Set<ConfusionParameter> set = mismatchesOnSeq.get(index);
			if (set.isEmpty()) {
				/* mismatch not registered, register as other */
				final int i = mapCharToInt(observ.iterator.current());
				final int j = mapCharToInt(predict.iterator.current());
				set.add(ConfusionParameter.OTHER);
				m.matrix[i][j]++;
			}
			observ.iterator.next();
			predict.iterator.next();
		}
	}

	private void registerMismatch(final ConfusionParameter type, final int count) {

		/* rewind iterators */
		observ.iterator.setIndex(observ.iterator.getIndex() - count);
		predict.iterator.setIndex(predict.iterator.getIndex() - count);

		/* update counts */
		final ConfusionCount m = confusionMatrix.get(type);

		/* count sequences in confusion matrix */
		for (int x = 0, index = observ.iterator.getIndex(); x < count; x++, index++) {
			final int i = mapCharToInt(observ.iterator.current());
			final int j = mapCharToInt(predict.iterator.current());

			mismatchesOnSeq.get(index).add(type);

			m.matrix[i][j]++;
			observ.iterator.next();
			predict.iterator.next();
		}
	}

	/**
	 * Returns results for the State machine analysis.
	 * 
	 * @see dk.sdu.imada.jkkn04.Protein.tests.AbstractEvaluation#getResults()
	 */
	@Override
	public Map<String, Object> getResults() {
		final Map<String, Object> results = super.getResults();

		/**
		 * We calculate a total sum, this is weighted as:<br/><br/>
		 * 
		 * <pre>
		 * totalSum = (Bridges * 1 + Gaps *
		 * 1 + Expansions * 0.5 + Reductions * 0.5 + Other * 1).
		 * </pre>
		 * 
		 * A percentage is calculated:
		 * 
		 * <pre>
		 * 1 - (totalSum / sequenceLength) * 100
		 * </pre>
		 */
		double totalSum = 0;

		for (final Map.Entry<ConfusionParameter, ConfusionCount> e : confusionMatrix
				.entrySet()) {
			final ConfusionParameter param = e.getKey();
			final ConfusionCount count = e.getValue();
			int sum = 0;
			for (final State iState : State.values()) {
				final int i = iState.getInt();
				int sumOnObserved = 0;
				int sumOnPredicted = 0;
				for (final State jState : State.values()) {
					final int j = jState.getInt();
					results.put("State-" + iState.getChar() + jState.getChar() + "-"
							+ param.name(), count.matrix[i][j]);
					sum += count.matrix[i][j];
					sumOnObserved += count.matrix[i][j];
					sumOnPredicted += count.matrix[j][i];
				}
				results.put(
						"State-Sum-" + iState.getChar() + "-Observed-" + param.name(),
						sumOnObserved);
				results.put("State-Sum-" + iState.getChar() + "-Predicted-"
						+ param.name(), sumOnPredicted);

				/*
				 * Summing by expansion/bridges on the predicted state and reduction/gaps
				 * on the observed state
				 */
				switch (param) {
				case BRIDGE:
				case EXPANSION:
					results.put("State-Sum-" + iState.getChar() + "-" + param.name(),
							sumOnPredicted);
					break;
				case GAP:
				case REDUCTION:
					results.put("State-Sum-" + iState.getChar() + "-" + param.name(),
							sumOnObserved);
					break;
				case OTHER:
					break;
				}
			}
			results.put("State-Sum-" + param.name(), sum);
			totalSum += sum
					* ((param == ConfusionParameter.EXPANSION || param == ConfusionParameter.REDUCTION) ? 0.5
							: 1);
		}

		results.put("State-TotalSum", totalSum);
		results.put("State", 100 * (1 - totalSum / sequenceLength));
		results.put("length", sequenceLength);
		return results;
	}

	/**
	 * Sum with another State Machine result
	 * 
	 * @see dk.sdu.imada.jkkn04.Protein.tests.AbstractEvaluation#sumWith(dk.sdu.imada.jkkn04.Protein.tests.AbstractEvaluation)
	 */
	@Override
	public void sumWith(final AbstractEvaluation evaluation) {
		super.sumWith(evaluation);
		if (evaluation != null) {
			if (!(evaluation instanceof StateMachine)) {
				// FIXME: throw something
			} else {
				final StateMachine other = (StateMachine) evaluation;

				/* Sum confusion matrixes */
				for (final Map.Entry<ConfusionParameter, ConfusionCount> e : other.confusionMatrix
						.entrySet()) {
					final ConfusionParameter param = e.getKey();
					final ConfusionCount otherCount = e.getValue();
					final ConfusionCount count = confusionMatrix.get(param);
					for (final State iState : State.values()) {
						final int i = iState.getInt();
						for (final State jState : State.values()) {
							final int j = jState.getInt();
							count.matrix[i][j] += otherCount.matrix[i][j];
						}
					}
				}
				sequenceLength += other.sequenceLength;
			}
		}
	}

	/**
	 * Returns two strings representing where we found the gaps, extensions, bridges,
	 * reductions,...
	 * 
	 * @return two strings
	 */
	public String[] getMismatchesBySequence() {
		final StringBuffer strBridgeExpansionOther = new StringBuffer();
		final StringBuffer strGapReduction = new StringBuffer();

		if (sequences == 1) {
			// only valid if representing a single sequence
			for (int i = 0; i < sequenceLength; i++) {
				final Set<ConfusionParameter> atPos = mismatchesOnSeq.get(i);
				if (atPos.contains(ConfusionParameter.BRIDGE)) {
					strBridgeExpansionOther.append("B");
				} else if (atPos.contains(ConfusionParameter.EXPANSION)) {
					strBridgeExpansionOther.append("E");
				} else if (atPos.contains(ConfusionParameter.OTHER)) {
					strBridgeExpansionOther.append("O");
				} else {
					strBridgeExpansionOther.append(" ");
				}
				if (atPos.contains(ConfusionParameter.GAP)) {
					strGapReduction.append("G");
				} else if (atPos.contains(ConfusionParameter.REDUCTION)) {
					strGapReduction.append("R");
				} else {
					strGapReduction.append(" ");
				}
			}
		}
		return new String[] { strBridgeExpansionOther.toString(),
				strGapReduction.toString() };
	}

	/**
	 * Prints nice debug of where we found the mismatches.
	 */
	public void printDebug() {
		for (final ConfusionParameter p : ConfusionParameter.values()) {
			final char type = p.name().toLowerCase().charAt(0);
			for (int i = 0; i < sequenceLength; i++) {
				if (mismatchesOnSeq.get(i).contains(p)) {
					System.out.print(type);
				} else {
					System.out.print(" ");
				}
			}
			System.out.println();
		}
		System.out.println(observedProtein.getCleanSequence());
		System.out.println(predictedProtein.getCleanSequence());
	}

}
