package dk.sdu.imada.jkkn04.Protein.util;

import java.util.EnumSet;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import dk.sdu.imada.jkkn04.Protein.data.Protein;
import dk.sdu.imada.jkkn04.Protein.data.ProteinMap;
import dk.sdu.imada.jkkn04.Protein.data.SequenceTranslated;
import dk.sdu.imada.jkkn04.Protein.data.SequenceType;
import dk.sdu.imada.jkkn04.Protein.tests.AbstractEvaluation;
import dk.sdu.imada.jkkn04.Protein.tests.MCC;
import dk.sdu.imada.jkkn04.Protein.tests.QScore;
import dk.sdu.imada.jkkn04.Protein.tests.SOV;
import dk.sdu.imada.jkkn04.Protein.tests.StateMachine;

/**
 * Utility class to do processing of various kinds
 * 
 * @author Kristian Krmmer Nielsen <jkkn@jkkn.dk>
 */
public class ProcessUtil {

	/**
	 * Class for representing multiple evaluations order by evaluation type
	 * 
	 * @author Kristian Krmmer Nielsen <jkkn@jkkn.dk>
	 */
	public static class MultipleEvaluationsMap extends
			HashMap<String, AbstractEvaluation> {
		private static final long serialVersionUID = 1L;
	};

	/**
	 * Class for representing multiple evaluations ordered by observed and predicted
	 * sequence pair.
	 * 
	 * @author Kristian Krmmer Nielsen <jkkn@jkkn.dk>
	 */
	public static class ObservationPredictionMap extends
			HashMap<SequenceTranslated, Map<SequenceTranslated, MultipleEvaluationsMap>> {
		private static final long serialVersionUID = 1L;
	};

	/**
	 * Set of all known evaluators in this implementation
	 */
	public static final AbstractEvaluation[] ALL_EVALUATIONS = { new QScore(), new SOV(),
			new MCC(), new StateMachine() };

	/**
	 * Method for performing all tests of a certain type
	 * 
	 * @param pm Map of one protein, multiple sequence types
	 * @param observationTypes Observed translated types to use
	 * @param predictionTypes Predicted translated types to use
	 * @param evaluationClasses Evaluator classes to use (can be ALL_EVALUATIONS)
	 * @return map of all results for the given protein
	 */
	public static ObservationPredictionMap performAllTests(final ProteinMap pm,
			final Set<SequenceTranslated> observationTypes,
			final Set<SequenceTranslated> predictionTypes,
			final AbstractEvaluation[] evaluationClasses) {

		final ObservationPredictionMap apm = new ObservationPredictionMap();

		/*
		 * go through every combination of observations available to predictions available
		 */
		for (final SequenceTranslated observationType : observationTypes) {
			if (pm.containsKey(observationType)) {
				final Protein observed = pm.get(observationType);
				final Map<SequenceTranslated, MultipleEvaluationsMap> observedMap = new HashMap<SequenceTranslated, MultipleEvaluationsMap>();

				for (final SequenceTranslated predictionType : predictionTypes) {
					if (pm.containsKey(predictionType)) {
						final MultipleEvaluationsMap multiEvaluationsMap = new MultipleEvaluationsMap();
						final Protein predicition = pm.get(predictionType);
						for (final AbstractEvaluation eval : evaluationClasses) {
							try {
								final AbstractEvaluation current = eval.clone();
								current.setObservedProtein(observed);
								current.setPredictedProtein(predicition);
								current.performEvaluation();

								multiEvaluationsMap.put(current.getClass().getName(),
										current);

							} catch (final IllegalArgumentException e) {
								e.printStackTrace();
							} catch (final SecurityException e) {
								e.printStackTrace();
							}
						}
						observedMap.put(predictionType, multiEvaluationsMap);
					}
				}

				if (!observedMap.isEmpty()) {
					apm.put(observationType, observedMap);
				}
			}
		}

		return apm;
	}

	/**
	 * This summarises results for a protein into total sum.
	 * 
	 * @param sum Map-object to use for summarisation (can be an empty map)
	 * @param map Object representing a result, which should be summed into <code>sum</code>.
	 */
	public static void summerizeAll(final ObservationPredictionMap sum,
			final ObservationPredictionMap map) {

		for (final Map.Entry<SequenceTranslated, Map<SequenceTranslated, MultipleEvaluationsMap>> observed : map
				.entrySet()) {
			final SequenceTranslated observedType = observed.getKey();
			Map<SequenceTranslated, MultipleEvaluationsMap> sumObserved = sum
					.get(observedType);
			if (sumObserved == null) {
				sum
						.put(
								observedType,
								sumObserved = new HashMap<SequenceTranslated, MultipleEvaluationsMap>());
			}
			for (final Map.Entry<SequenceTranslated, MultipleEvaluationsMap> prediction : observed
					.getValue().entrySet()) {
				final SequenceTranslated predictionType = prediction.getKey();
				MultipleEvaluationsMap sumPrediction = sumObserved.get(predictionType);
				if (sumPrediction == null) {
					sumObserved.put(predictionType,
							sumPrediction = new MultipleEvaluationsMap());
				}
				for (final Map.Entry<String, AbstractEvaluation> evaluation : prediction
						.getValue().entrySet()) {
					final AbstractEvaluation sumOfEvaluation = sumPrediction
							.get(evaluation.getKey());
					if (sumOfEvaluation == null) {
						sumPrediction.put(evaluation.getKey(), evaluation.getValue()
								.clone());
					} else {
						sumOfEvaluation.sumWith(evaluation.getValue());
					}
				}
			}
		}

	}

	/**
	 * Shortcut to perform all evaluation by translated types
	 * 
	 * @param pm Map of multiple sequences types of a protein
	 * @param observationTypes Observed translated types to use
	 * @param predictionTypes Prediction translated types to use
	 * @return result map ordered by translated types
	 */
	public static ObservationPredictionMap performAllTests(final ProteinMap pm,
			final Set<SequenceTranslated> observationTypes,
			final Set<SequenceTranslated> predictionTypes) {
		return performAllTests(pm, observationTypes, predictionTypes, ALL_EVALUATIONS);
	}

	/**
	 * Shortcut to perform all evaluation by type
	 * 
	 * @param pm Map of multiple sequences types of a protein
	 * @param observationTypes Observed types to use
	 * @param predictionTypes Prediction types to use
	 * @return result map ordered by translated types
	 */
	public static ObservationPredictionMap performAllTestsByType(final ProteinMap pm,
			final Set<SequenceType> observationTypes,
			final Set<SequenceType> predictionTypes) {
		return performAllTests(pm, SequenceTranslated.buildAllOf(observationTypes),
				SequenceTranslated.buildAllOf(predictionTypes));
	}

	/**
	 * Shortcut to perform all evaluation by type
	 * 
	 * @param pm Map of multiple sequences types of a protein
	 * @param observationTypes Observed types to use
	 * @param predictionTypes Prediction types to use
	 * @param evaluationClasses Evaluator classes to use (can be ALL_EVALUATIONS)
	 * @return result map ordered by translated types
	 */
	public static ObservationPredictionMap performAllTestsByType(final ProteinMap pm,
			final EnumSet<SequenceType> observationTypes,
			final EnumSet<SequenceType> predictionTypes,
			final AbstractEvaluation[] evaluationClasses) {
		return performAllTests(pm, SequenceTranslated.buildAllOf(observationTypes),
				SequenceTranslated.buildAllOf(predictionTypes), evaluationClasses);
	}

}
